package togos.compositor;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;

import javax.imageio.ImageIO;

public class Compositor
{
	public final int w, h;
	public final float[] intensities;
	public final int[][] buffers;
	public final int[] outBuffer;
	
	public Compositor( int w, int h, int bufferCount ) {
		this.w = w; this.h = h;
		this.outBuffer = new int[w*h];
		this.buffers = new int[bufferCount][w*h];
		this.intensities = new float[bufferCount*3];
	}
	
	public void setColor( int bufferId, float r, float g, float b ) {
		assert bufferId >= 0;
		assert bufferId < buffers.length;
		intensities[bufferId*3+0] = r;
		intensities[bufferId*3+1] = g;
		intensities[bufferId*3+2] = b;
	}
	
	public void loadImage( int bufferId, File file ) throws IOException {
		assert bufferId >= 0;
		assert bufferId < buffers.length;
		BufferedImage img = ImageIO.read(file);
		img.getRGB(0,  0, w, h, buffers[bufferId], 0, w );
	}
	
	protected static final int clamp255( float i ) {
		if( i <   0 ) return 0;
		if( i > 255 ) return 255;
		return (int)i;
	}
	
	protected static final int add( int orig, int add, float rFact, float gFact, float bFact ) {
		float r = ((orig >> 16)&0xFF) + ((add >> 16)&0xFF)*rFact;
		float g = ((orig >>  8)&0xFF) + ((add >>  8)&0xFF)*gFact;
		float b = ((orig >>  0)&0xFF) + ((add >>  0)&0xFF)*bFact;
		
		float r2 = r, g2 = g, b2 = b;
		// Allow individual RGB components to brighten the pixel towards white
		// if they are high enough:
		if( r > 255 ) {
			g2 += (r-255)/2;
			b2 += (r-255)/2;
		}
		if( g > 255 ) {
			r2 += (g-255)/2;
			b2 += (g-255)/2;
		}
		if( b > 255 ) {
			r2 += (b-255)/2;
			g2 += (b-255)/2;
		}
		
		return 0xFF000000 | (clamp255(r2) << 16) | (clamp255(g2) << 8) | (clamp255(b2) << 0);
	}
	
	public void recompose() {
		for( int i=w*h-1; i>=0; --i ) outBuffer[i] = 0xFF000000;
		
		for( int l=buffers.length-1; l>=0; --l ) {
			float rFact = intensities[l*3+0];
			float gFact = intensities[l*3+1];
			float bFact = intensities[l*3+2];
			if( rFact == 0 && gFact == 0 & bFact == 0 ) continue;
			
			for( int i=w*h-1; i>=0; --i ) {
				outBuffer[i] = add( outBuffer[i], buffers[l][i], rFact, gFact, bFact );
			}
		}
	}
	
	public void toBufferedImage( BufferedImage bi ) {
		int minW = Math.min(w, bi.getWidth());
		int minH = Math.min(h, bi.getHeight());
		bi.setRGB(0, 0, minW, minH, outBuffer, 0, w);
	}
	
	protected static final int cycle( int current, int max ) {
		current += 1;
		return ( current >= max ) ? 0 : current; 
	}
	
	protected static final float override( int state, float onVal, float defaultVal ) {
		if( state == 0 ) return 0.0f;
		if( state == 1 ) return onVal;
		return defaultVal;
	}
	
	public static final String USAGE =
		"Usage: compositor [options] <image1> <image2> ...\n" +
		"Options:\n" +
		"  -repaint-interval <milliseconds> ; specify time between repaints\n" +
		"  -disco                           ; gives lights different colors\n" +
		"  -trippy                          ; allows negative colors\n" +
		"\n" +
		"Keyboard controls:\n" +
		"  0-9 ; cycle modes of individual lights\n" +
		"  T   ; toggle 'trippy mode'\n" +
		"  D   ; toggle 'disco mode'\n";
	
	static volatile boolean goah = true;
	static volatile boolean trippy = false;	
	static volatile long repaintInterval = 0;
	static volatile boolean disco = false;
	
	public static void main( String[] args ) throws Exception {
		ArrayList<String> imageFilenames = new ArrayList<String>();
		
		for( int i=0; i<args.length; ++i ) {
			if( "-trippy".equals(args[i]) ) {
				trippy = true;
			} else if( "-disco".equals(args[i]) ) {
				disco = true;
			} else if( "-repaint-interval".equals(args[i]) ) {
				repaintInterval = Long.parseLong(args[++i]);
			} else if( "-h".equals(args[i]) || "-help".equals(args[i]) || "-?".equals(args[i]) || "--help".equals(args[i]) ) {
				System.out.println(USAGE);
				System.exit(0);
			} else if( !args[i].startsWith("-") ) {
				imageFilenames.add(args[i]);
			} else {
				System.err.println("Error: Unrecognized argument: "+args[i]);
				System.err.println(USAGE);
				System.exit(1);
			}
		}
		
		int imageCount = imageFilenames.size();
		if( imageCount == 0 ) {
			System.err.println("Error: You need ta gimmeh some fiulz");
			System.err.println(USAGE);
			System.exit(1);
		}
		
		BufferedImage i0; // Used to determine canvas size 
		try {
			i0 = ImageIO.read(new File(imageFilenames.get(0)));
		} catch( IOException e ) {
			System.err.println("Error loading "+imageFilenames.get(0));
			throw e;
		}
		
		final int[] lightStates = new int[imageCount];
		final Compositor comp = new Compositor( i0.getWidth(), i0.getHeight(), imageCount );
		for( int i=0; i<imageCount; ++i ) {
			try {
				comp.loadImage(i, new File(imageFilenames.get(i)));
			} catch( IOException e ) {
				System.err.println("Error loading "+imageFilenames.get(i));
			}
			comp.setColor(i, 0.15f, 0.15f, 0.15f);
			lightStates[i] = 2;
		}
		comp.recompose();
		
		final BufferedImage bi = new BufferedImage(comp.w, comp.h, BufferedImage.TYPE_INT_ARGB);
		comp.toBufferedImage(bi);
		
		final Frame frame = new Frame("TESTME.1ST");
		final Canvas canv = new Canvas() {
			private static final long serialVersionUID = 1L;
			public void paint( Graphics g ) {
				synchronized( bi ) { g.drawImage(bi, 0, 0, null); }
			}
			public void update( Graphics g ) { paint(g); }
		};
		canv.setPreferredSize(new Dimension(comp.w,comp.h));
		frame.add(canv);
		frame.pack();
		frame.addWindowListener(new WindowAdapter() {
			public void windowClosing( WindowEvent wEvt ) {
				frame.dispose();
				goah = false;
			}
		});
		frame.setBackground(Color.BLACK);
		frame.setVisible(true);
		canv.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				switch( e.getKeyCode() ) {
				case KeyEvent.VK_T:
					trippy ^= true; break;
				case KeyEvent.VK_D:
					disco ^= true; break;
				default:
					if( e.getKeyCode() >= KeyEvent.VK_0 && e.getKeyCode() <= KeyEvent.VK_9 ) {
						int layerId = e.getKeyCode() - KeyEvent.VK_0;
						if( layerId < 0 || layerId >= lightStates.length ) return;
						lightStates[layerId] = cycle(lightStates[layerId], 3);
					}
				}
			}
		});
		canv.requestFocus();
		
		new Thread() {
			@Override
			public void run() {
				long i = 0; 

				final float generalIntensityMultiplier = Math.min( 1, 2f / (comp.buffers.length - 2) ); 
				
				while( goah ) {
					try {
						Thread.sleep(repaintInterval);
					} catch( InterruptedException e ) {
						return;
					}
					
					final float sinMult = trippy ? 1.0f : 0.5f;
					final float sinAdd  = trippy ? 0.0f : 0.5f;
					
					for( int j=0; j<comp.buffers.length; ++j ) {
						float intensity = generalIntensityMultiplier * override( lightStates[j], 1.0f, sinAdd + sinMult*(float)Math.sin( i * (j+2) * 0.01 ) );
						float rFact, gFact, bFact;
						if( disco ) {
							rFact = (j%3 == 0) ? 1.0f : 0.5f;
							gFact = (j%3 == 1) ? 1.0f : 0.5f;
							bFact = (j%3 == 2) ? 1.0f : 0.5f;
						} else {
							rFact = gFact = bFact = 1.0f;
						}
						comp.setColor( j, rFact*intensity, gFact*intensity, bFact*intensity );
					}
					
					comp.recompose();
					synchronized( bi ) { comp.toBufferedImage(bi); }
					canv.repaint();
					
					++i;
				}
			}
		}.start();
	}
}
